@@ -8,6 +8,8 @@ import time
 import uuid
 import requests
 import traceback
+import os
+import hashlib
 
 from ..config import NERD_COMPUTE_ENDPOINT, DEBUG_MODE
 from ..spinner import Spinner
@@ -20,6 +22,264 @@ from .auth import get_api_key
 from .storage.storage import upload_nerd_cloud_storage
 from .on_server.on_server import generate_cloud_code, deserialize_arg
 
+# Data cache configuration
+CACHE_DIR = os.path.join(os.path.expanduser("~"), ".nerd_compute", "data_cache")
+CACHE_FILE = os.path.join(CACHE_DIR, "data_fingerprints.json")
+os.makedirs(CACHE_DIR, exist_ok=True)
+
+def load_data_cache():
+    """Load the data fingerprint cache"""
+    if os.path.exists(CACHE_FILE):
+        try:
+            with open(CACHE_FILE, 'r') as f:
+                return json.load(f)
+        except Exception as e:
+            debug_print(f"Error loading data cache: {e}")
+    return {}
+
+def save_data_cache(cache):
+    """Save the data fingerprint cache"""
+    try:
+        with open(CACHE_FILE, 'w') as f:
+            json.dump(cache, f)
+    except Exception as e:
+        debug_print(f"Error saving data cache: {e}")
+
+def compute_data_fingerprint(data):
+    """
+    Compute a unique fingerprint for data to check for duplicates.
+    Uses sampling approach for large data.
+    
+    Args:
+        data: The data to fingerprint
+        
+    Returns:
+        str: SHA-256 hexdigest fingerprint or None if fingerprinting fails
+    """
+    try:
+        import numpy as np
+        
+        # Special handling for numpy arrays (common in astronomy data)
+        if isinstance(data, np.ndarray):
+            h = hashlib.sha256()
+            # Include shape and dtype in fingerprint
+            h.update(str(data.shape).encode())
+            h.update(str(data.dtype).encode())
+            
+            # For large arrays, sample values instead of using whole array
+            if data.size > 1_000_000:  # 1 million elements
+                sample_size = 100_000  # Sample 100k elements
+                
+                # Sample start, middle, and end of array
+                h.update(data.flat[:sample_size].tobytes())
+                mid_point = data.size // 2
+                h.update(data.flat[mid_point:mid_point+sample_size].tobytes())
+                h.update(data.flat[-sample_size:].tobytes())
+            else:
+                h.update(data.tobytes())
+            return h.hexdigest()
+            
+        # For bytes objects
+        elif isinstance(data, bytes):
+            h = hashlib.sha256()
+            # Use data length as part of fingerprint
+            data_len = len(data)
+            h.update(str(data_len).encode())
+            
+            # Sample data for very large objects
+            if data_len > 10*1024*1024:  # 10MB
+                sample_size = 1024*1024  # 1MB samples
+                # First 1MB
+                h.update(data[:sample_size])
+                # Middle 1MB
+                middle_pos = data_len // 2
+                h.update(data[middle_pos:middle_pos+sample_size])
+                # Last 1MB
+                h.update(data[-sample_size:])
+            else:
+                h.update(data)
+            return h.hexdigest()
+            
+        # For other objects, pickle first then fingerprint
+        # This handles pandas DataFrames, astropy Tables, etc.
+        pickled = pickle.dumps(data, protocol=4)
+        return compute_data_fingerprint(pickled)
+        
+    except Exception as e:
+        debug_print(f"Error computing fingerprint: {e}")
+        return None
+
+def get_cached_data_reference(fingerprint):
+    """
+    Check if data with the given fingerprint exists in cache
+    
+    Args:
+        fingerprint: The data fingerprint to look up
+        
+    Returns:
+        dict: The reference object if found, None otherwise
+    """
+    if not fingerprint:
+        return None
+        
+    data_cache = load_data_cache()
+    if fingerprint in data_cache:
+        cached_info = data_cache[fingerprint]
+        debug_print(f"Found cached data: {cached_info['dataId']}")
+        
+        # Update last access time
+        data_cache[fingerprint]['lastAccessed'] = time.strftime('%Y-%m-%d %H:%M:%S')
+        save_data_cache(data_cache)
+        
+        return {
+            'type': 'bytes_reference',
+            'value': {
+                'data_reference': cached_info['dataId'],
+                's3Uri': cached_info.get('s3Uri', ''),
+                'sizeMB': cached_info.get('sizeMB', '')
+            }
+        }
+    return None
+
+def cache_data_reference(fingerprint, upload_result):
+    """
+    Store data reference in the fingerprint cache
+    
+    Args:
+        fingerprint: The data fingerprint
+        upload_result: The result from upload_nerd_cloud_storage
+    """
+    if not fingerprint:
+        return
+        
+    data_cache = load_data_cache()
+    data_cache[fingerprint] = {
+        'dataId': upload_result['dataId'],
+        's3Uri': upload_result.get('s3Uri', ''),
+        'sizeMB': upload_result.get('sizeMB', ''),
+        'uploadDate': time.strftime('%Y-%m-%d %H:%M:%S'),
+        'lastAccessed': time.strftime('%Y-%m-%d %H:%M:%S')
+    }
+    save_data_cache(data_cache)
+    debug_print(f"Cached data reference: {upload_result['dataId']} with fingerprint: {fingerprint}")
+
+def serialize_with_chunking(obj):
+    try:
+        # Special handling for large binary data that should go to S3 directly
+        if isinstance(obj, dict) and "__nerd_data_reference" in obj:
+            debug_print(f"Found reference object, serializing reference only")
+            return {
+                'type': 'bytes_reference',
+                'value': {
+                    'data_reference': obj['__nerd_data_reference'],
+                    's3Uri': obj.get('__nerd_s3_uri', ''),
+                    'sizeMB': obj.get('__nerd_size_mb', '')
+                }
+            }
+            
+        # For large data that might get uploaded to cloud storage,
+        # check if we've already uploaded identical data before
+        if is_large_data(obj, threshold_mb=3):
+            debug_print("Large data detected, computing fingerprint...")
+            
+            # Compute data fingerprint for deduplication
+            fingerprint = compute_data_fingerprint(obj)
+            
+            # Check if we've already uploaded this exact data
+            if fingerprint:
+                cached_reference = get_cached_data_reference(fingerprint)
+                if cached_reference:
+                    debug_print(f"Found identical data already in cloud storage!")
+                    return cached_reference
+                    
+            # Data not found in cache, proceed with upload
+            debug_print("Data not in cache, uploading...")
+            upload_result = upload_nerd_cloud_storage(obj)
+            
+            # Cache the result for future reference
+            if fingerprint:
+                cache_data_reference(fingerprint, upload_result)
+                
+            return {
+                'type': 'bytes_reference',
+                'value': {
+                    'data_reference': upload_result['dataId'],
+                    's3Uri': upload_result.get('s3Uri', ''),
+                    'sizeMB': upload_result.get('sizeMB', '')
+                }
+            }
+
+        # Handle functions or callable objects differently
+        if callable(obj):
+            # Special case for partial objects (common in parallel processing)
+            if isinstance(obj, functools.partial):
+                debug_print(f"Serializing partial function: {obj.func.__name__}")
+                return {
+                    'type': 'callable',
+                    'function_type': 'partial',
+                    'func_name': obj.func.__name__,
+                    'func_module': obj.func.__module__,
+                    'args': serialize_with_chunking(obj.args),
+                    'keywords': serialize_with_chunking(obj.keywords)
+                }
+            # For regular functions
+            else:
+                func_name = getattr(obj, '__name__', 'unknown_function')
+                func_module = getattr(obj, '__module__', '__main__')
+                debug_print(f"Serializing function: {func_name} from module {func_module}")
+                return {
+                    'type': 'callable',
+                    'function_type': 'function',
+                    'func_name': func_name,
+                    'func_module': func_module,
+                }
+
+        # Standard serialization for everything else:
+        # 1. Pickle the object to binary form
+        pickled = pickle.dumps(obj)
+        # 2. Compress the pickled data
+        compressed = zlib.compress(pickled)
+        # 3. Base64 encode the compressed data
+        encoded = base64.b64encode(compressed).decode('utf-8')
+        debug_print(f"Serialized object size: {len(encoded) / (1024*1024):.2f} MB")
+        return {'type': 'data', 'value': encoded}
+    except Exception as e:
+        spinner.stop()
+        print(f"⚠️ Warning: Could not package object: {e}")
+        spinner.start()
+        return {'type': 'string', 'value': str(obj)}
+
+# Add functions to manage the cache from outside the module
+def list_cached_data():
+    """List all data fingerprints in the cache"""
+    cache = load_data_cache()
+    if not cache:
+        print("ℹ️ No cached data fingerprints found")
+        return
+    
+    print(f"Found {len(cache)} cached data entries:")
+    for fingerprint, info in cache.items():
+        size = info.get('sizeMB', 'unknown size')
+        upload_date = info.get('uploadDate', 'unknown date')
+        last_accessed = info.get('lastAccessed', 'unknown')
+        data_id = info.get('dataId', 'unknown ID')
+        print(f"- Data ID: {data_id}")
+        print(f"  Size: {size} MB")
+        print(f"  Uploaded: {upload_date}")
+        print(f"  Last accessed: {last_accessed}")
+        print()
+
+def clear_data_cache():
+    """Clear the data fingerprint cache"""
+    if os.path.exists(CACHE_FILE):
+        try:
+            os.remove(CACHE_FILE)
+            print("✅ Data fingerprint cache cleared")
+        except Exception as e:
+            print(f"❌ Error clearing cache: {e}")
+    else:
+        print("ℹ️ Cache file doesn't exist")
+
 def cloud_compute(cores=8, timeout=1800):
     """
     A special function decorator that sends your computation to a powerful cloud server.
@@ -188,35 +448,6 @@ def cloud_compute(cores=8, timeout=1800):
             debug_print(f"Generated import block:\n{import_block}")
 
             # Serialize the arguments for cloud execution
-            def serialize_with_chunking(obj):
-                try:
-                    # Special handling for large binary data that should go to S3 directly
-                    if isinstance(obj, dict) and "__nerd_data_reference" in obj:
-                        debug_print(f"Found reference object, serializing reference only")
-                        return {
-                            'type': 'bytes_reference',
-                            'value': {
-                                'data_reference': obj['__nerd_data_reference'],
-                                's3Uri': obj.get('__nerd_s3_uri', ''),
-                                'sizeMB': obj.get('__nerd_size_mb', '')
-                            }
-                        }
-
-                    # Standard serialization for everything else:
-                    # 1. Pickle the object to binary form
-                    pickled = pickle.dumps(obj)
-                    # 2. Compress the pickled data
-                    compressed = zlib.compress(pickled)
-                    # 3. Base64 encode the compressed data
-                    encoded = base64.b64encode(compressed).decode('utf-8')
-                    debug_print(f"Serialized object size: {len(encoded) / (1024*1024):.2f} MB")
-                    return {'type': 'data', 'value': encoded}
-                except Exception as e:
-                    spinner.stop()
-                    print(f"⚠️ Warning: Could not package object: {e}")
-                    spinner.start()
-                    return {'type': 'string', 'value': str(obj)}
-
             serialized_args = []
             for arg in args:
                 serialized_args.append(serialize_with_chunking(arg))
