@@ -385,6 +385,39 @@ def auto_reference_wrapper(func, args, kwargs):
     # Call the function with resolved arguments
     print(f"[REFERENCE_DEBUG] Calling function {func.__name__} with resolved arguments")
     try:
+        # Special handling for grab_noise which likely uses multiprocessing inside
+        if func.__name__ == 'grab_noise':
+            # Import the function to safely wrap non-callable objects as callable
+            try:
+                from nerd_mega_compute.cloud.storage.data_parsing_utils import ensure_callable
+                
+                # Check the arguments for potential multiprocessing functions
+                for i, arg in enumerate(resolved_args):
+                    # Look for dictionaries that should be functions or other non-callable objects
+                    # that might be passed to multiprocessing
+                    if isinstance(arg, dict) and (
+                            arg.get('type') == 'callable' or 
+                            'func_name' in arg or 
+                            'function_type' in arg):
+                        print(f"[REFERENCE_DEBUG] Found potential callable dictionary in arg[{i}], wrapping it")
+                        resolved_args[i] = ensure_callable(arg)
+                    elif isinstance(arg, str) and arg.startswith("<function ") and "at 0x" in arg:
+                        print(f"[REFERENCE_DEBUG] Found function string in arg[{i}], wrapping it")
+                        resolved_args[i] = ensure_callable(arg)
+                        
+                # Also check for dictionaries inside first arg that might need conversion
+                if resolved_args and isinstance(resolved_args[0], dict):
+                    for k, v in resolved_args[0].items():
+                        if (isinstance(v, dict) and (v.get('type') == 'callable' or 'func_name' in v)) or \
+                           (isinstance(v, str) and v.startswith("<function ") and "at 0x" in v):
+                            print(f"[REFERENCE_DEBUG] Found potential callable in arg[0][{k}], wrapping it")
+                            resolved_args[0][k] = ensure_callable(v, k)
+            except Exception as e:
+                print(f"[REFERENCE_DEBUG] Error ensuring callables: {e}")
+                import traceback
+                traceback.print_exc()
+        
+        # Call the function with the resolved and fixed arguments
         result = func(*resolved_args, **resolved_kwargs)
         print(f"[REFERENCE_DEBUG] Function {func.__name__} executed successfully, result type: {type(result).__name__}")
         return result
